@@ -5,6 +5,7 @@ from geomdl.visualization import VisMPL
 import numpy as np
 from copy import deepcopy
 from .optimize_connection_v3 import *
+from .optimize_connection_v4 import *
 from tqdm import trange
 import matplotlib.pyplot as plt
 from mpl_toolkits.mplot3d import Axes3D
@@ -23,6 +24,122 @@ def get_initial_vectors(forest,network,assignment,tree,edge):
     V2 = V2/np.linalg.norm(V2)
     return V1,V2,P4
 
+def link_v2(forest,network_id,tree_idx,tree_jdx,radius_buffer=0):
+    """
+    Method for connecting vascular trees together
+
+    Parameters
+    ----------
+                forest : svcco.forest object
+                       the forest object containing at least two trees to be
+                       connected
+                network_id : int
+                       the index of the network containing two trees that will be
+                       connected
+                tree_idx : int
+                       the index of the first tree in the network for connecting
+                tree_jdx : int
+                       the index of the second tree in the network for connecting
+                radius_buffer : float
+                       the minimum distance between the connecting vessels and any
+                       other vessel within the network
+
+    Returns
+    -------
+                connections : list of connection objects
+                        for each connection between terminals in a network a
+                        connection object is generated and solved
+    """
+    number_of_connections = forest.assignments[network_id][tree_idx].shape[0]
+    connections = []
+    assign_idx = forest.assignments[network_id][tree_idx]
+    assign_jdx = forest.assignments[network_id][tree_jdx]
+    print('Linking Network {}, Tree {} -> Tree {}'.format(network_id,tree_idx,tree_jdx))
+    for cdx in trange(number_of_connections):
+        conn = connection(forest,network_id,tree_idx,tree_jdx,cdx,radius_buffer)
+        for pconn in connections:
+            conn.add_collision_vessels(pconn)
+        conn.solve()
+        conn.build_vessels()
+        connections.append(conn)
+    tree_idx_connections = []
+    tree_jdx_connections = []
+    count_idx = 0
+    count_jdx = 0
+    for cdx in trange(number_of_connections):
+        # Replace the current terminal vessel with the first vessel of the connection
+        forest.networks[network_id][tree_idx].data[assign_idx[cdx],0:3]   = connections[cdx].vessels[0][0,0:3]
+        forest.networks[network_id][tree_idx].data[assign_idx[cdx],3:6]   = connections[cdx].vessels[0][0,3:6]
+        forest.networks[network_id][tree_idx].data[assign_idx[cdx],12:15] = connections[cdx].vessels[0][0,12:15]
+        forest.networks[network_id][tree_idx].data[assign_idx[cdx],20]    = connections[cdx].vessels[0][0,20]
+        connections[cdx].vessels[0][0,21] = forest.networks[network_id][tree_idx].data[assign_idx[cdx],21]
+        connections[cdx].vessels[0][0,22] = forest.networks[network_id][tree_idx].data[assign_idx[cdx],22]
+        connections[cdx].vessels[0][0,26] = forest.networks[network_id][tree_idx].data[assign_idx[cdx],26]
+        connections[cdx].vessels[0][0,28] = forest.networks[network_id][tree_idx].data[assign_idx[cdx],28]
+        parent_vessel_index = int(assign_idx[cdx])
+        next_vessel_index   = int(max(forest.networks[network_id][tree_idx].data[:,-1]))+1+count_idx
+        previous_node_index = int(forest.networks[network_id][tree_idx].data[assign_idx[cdx],19])
+        next_node_index     = int(max(forest.networks[network_id][tree_idx].data[:,19]))+1
+        for vdx in range(1,connections[cdx].vessels[0].shape[0]):
+            connections[cdx].vessels[0][vdx,-1] = next_vessel_index
+            connections[cdx].vessels[0][vdx,17] = parent_vessel_index
+            connections[cdx].vessels[0][vdx,18] = previous_node_index
+            connections[cdx].vessels[0][vdx,19] = next_node_index
+            connections[cdx].vessels[0][vdx,21] = connections[cdx].vessels[0][vdx-1,21]
+            connections[cdx].vessels[0][vdx,22] = connections[cdx].vessels[0][vdx-1,22]
+            connections[cdx].vessels[0][vdx,26] = connections[cdx].vessels[0][vdx-1,26]+1
+            connections[cdx].vessels[0][vdx,28] = connections[cdx].vessels[0][vdx-1,28]
+            connections[cdx].vessels[0][vdx-1,15] = next_vessel_index
+            connections[cdx].vessels[0][vdx-1,23] = 1
+            connections[cdx].vessels[0][vdx-1,24] = 0
+            #forest.networks[network_id][tree_idx].data = np.vstack((forest.networks[network_id][tree_idx].data,connections[cdx].vessels[0][vdx,:].reshape(1,-1)))
+            parent_vessel_index = next_vessel_index
+            next_vessel_index += 1
+            previous_node_index = next_node_index
+            next_node_index += 1
+            count_idx += 1
+        tree_idx_connections.append(connections[cdx].vessels[0][1:,:])
+        # Replace the current terminal vessel with the first vessel of the connection
+        # for tree jdx
+        forest.networks[network_id][tree_jdx].data[assign_jdx[cdx],0:3]   = connections[cdx].vessels[1][0,0:3]
+        forest.networks[network_id][tree_jdx].data[assign_jdx[cdx],3:6]   = connections[cdx].vessels[1][0,3:6]
+        forest.networks[network_id][tree_jdx].data[assign_jdx[cdx],12:15] = connections[cdx].vessels[1][0,12:15]
+        forest.networks[network_id][tree_jdx].data[assign_jdx[cdx],20]    = connections[cdx].vessels[1][0,20]
+        connections[cdx].vessels[1][0,21] = forest.networks[network_id][tree_jdx].data[assign_jdx[cdx],21]
+        connections[cdx].vessels[1][0,22] = forest.networks[network_id][tree_jdx].data[assign_jdx[cdx],22]
+        connections[cdx].vessels[1][0,26] = forest.networks[network_id][tree_jdx].data[assign_jdx[cdx],26]
+        connections[cdx].vessels[1][0,28] = forest.networks[network_id][tree_jdx].data[assign_jdx[cdx],28]
+        parent_vessel_index = int(assign_jdx[cdx])
+        next_vessel_index   = int(max(forest.networks[network_id][tree_jdx].data[:,-1]))+1+count_jdx
+        previous_node_index = int(forest.networks[network_id][tree_jdx].data[assign_jdx[cdx],19])
+        next_node_index     = int(max(forest.networks[network_id][tree_jdx].data[:,19]))+1
+        for vdx in range(1,connections[cdx].vessels[1].shape[0]):
+            connections[cdx].vessels[1][vdx,-1] = next_vessel_index
+            connections[cdx].vessels[1][vdx,17] = parent_vessel_index
+            connections[cdx].vessels[1][vdx,18] = previous_node_index
+            connections[cdx].vessels[1][vdx,19] = next_node_index
+            connections[cdx].vessels[1][vdx,21] = connections[cdx].vessels[1][vdx-1,21]
+            connections[cdx].vessels[1][vdx,22] = connections[cdx].vessels[1][vdx-1,22]
+            connections[cdx].vessels[1][vdx,26] = connections[cdx].vessels[1][vdx-1,26]+1
+            connections[cdx].vessels[1][vdx,28] = connections[cdx].vessels[1][vdx-1,28]
+            connections[cdx].vessels[1][vdx-1,15] = next_vessel_index
+            connections[cdx].vessels[1][vdx-1,23] = 1
+            connections[cdx].vessels[1][vdx-1,24] = 0
+            #forest.networks[network_id][tree_jdx].data = np.vstack((forest.networks[network_id][tree_jdx].data,connections[cdx].vessels[1][vdx,:].reshape(1,-1)))
+            parent_vessel_index = next_vessel_index
+            next_vessel_index += 1
+            previous_node_index = next_node_index
+            next_node_index += 1
+            count_jdx += 1
+        tree_jdx_connections.append(connections[cdx].vessels[1][1:,:])
+    for cdx in range(1,len(tree_idx_connections)):
+        tree_idx_connections[0] = np.vstack((tree_idx_connections[0],tree_idx_connections[cdx]))
+    for cdx in range(1,len(tree_jdx_connections)):
+        tree_jdx_connections[0] = np.vstack((tree_jdx_connections[0],tree_jdx_connections[cdx]))
+    tree_idx_connections = tree_idx_connections[0]
+    tree_jdx_connections = tree_jdx_connections[0]
+    all_connections = [[tree_idx_connections,tree_jdx_connections]]
+    return all_connections, None, None
 
 def link(forest,radius_buffer=0):
     all_connections = []
@@ -393,10 +510,6 @@ def smooth(forest,curve_sample_size_min=5,curve_sample_size_max=21,curve_degree=
             connections.append([None])
     return connections, None, splines
 
-
-#def get_angle(V1,V2):
-#    return np.arccos(np.dot(V1,V2))*(180/np.pi)
-
 def smooth_new(forest,max_angle=45):
     connections = []
     for network in range(forest.number_of_networks):
