@@ -10,5 +10,5 @@ centerline ="""if not terminating:
     centerline = vmtk.centerlines(mesher.get_model_polydata(),[inlet],outlets,use_face_ids=True)
     centerline_model = modeling.PolyData()
     centerline_model.set_surface(centerline)
-    centerline_model.write('{}'+'/centerlines','vtp')
+    centerline_model.write('{}'+os.sep+'centerlines','vtp')
 """
